import textwrap

import pytest


def test_none_checked_var_return_transform(fix_text):
    new_text, _ = fix_text(
        """\
        if foo is None:
            return foo
        """
    )
    assert new_text == textwrap.dedent(
        """\
        if foo is None:
            return None
        """
    )


@pytest.mark.parametrize(
    "disable_comment",
    ("# fmt: off", "#fmt:off ", "#slyp: disable", "# slyp: disable=format "),
)
def test_none_checked_var_return_transform_disabled_with_fmt_off(
    fix_text, disable_comment
):
    fix_text(
        f"""\
        if foo is None:
            return foo{disable_comment}
        """,
        expect_changes=False,
    )


def test_none_checked_var_return_transform_preserves_leading_comment(fix_text):
    new_text, _ = fix_text(
        """\
        if foo is None:
            # explanation!
            return foo
        """
    )
    assert new_text == textwrap.dedent(
        """\
        if foo is None:
            # explanation!
            return None
        """
    )


def test_none_checked_var_return_transform_preserves_trailing_comment(fix_text):
    new_text, _ = fix_text(
        """\
        if foo is None:
            return foo
            # explanation!
        """
    )
    assert new_text == textwrap.dedent(
        """\
        if foo is None:
            return None
            # explanation!
        """
    )


def test_none_checked_var_return_handles_non_indented_block_case(fix_text):
    new_text, _ = fix_text(
        """\
        if foo is None: return foo
        """
    )
    assert new_text == textwrap.dedent(
        """\
        if foo is None:
            return None
        """
    )


@pytest.mark.parametrize("infix_whitespace", ["", " ", "  ", "   "])
def test_none_checked_var_return_preserves_trailing_comment_on_non_indented_block_case(
    fix_text, infix_whitespace
):
    new_text, _ = fix_text(
        f"""\
        if foo is None: return foo{infix_whitespace}# explainer!
        """
    )
    assert new_text == textwrap.dedent(
        """\
        if foo is None:  # explainer!
            return None
        """
    )



from __future__ import annotations

import argparse
import sys
import textwrap

from slyp.codes import CODE_MAP
from slyp.driver import driver_main


def main() -> None:
    parser = argparse.ArgumentParser(
        description="slyp is a linter and fixer for Python code",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "--list", action="store_true", help="list all error and warning codes"
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        help="increase output verbosity",
        default=0,
        dest="verbosity",
    )
    parser.add_argument(
        "--use-git-ls", action="store_true", help="find python files from git-ls-files"
    )
    parser.add_argument(
        "--only",
        choices=("fix", "lint"),
        help="Only fix or only lint.",
    )
    parser.add_argument(
        "--disable",
        help="Disable error and warning codes (comma delimited)",
        default="",
    )
    parser.add_argument(
        "--enable",
        help=(
            "Enable error and warning codes which are otherwise disabled "
            "(comma delimited, overrides --disable)"
        ),
        default="",
    )
    parser.add_argument(
        "--no-cache",
        help="Disable caching of results in the '.slyp_cache' directory.",
        action="store_true",
    )
    parser.add_argument("files", nargs="*", help="default: all python files")
    args = parser.parse_args()

    if args.list:
        list_codes()
        sys.exit(0)

    if args.use_git_ls and args.files:
        parser.error("--use-git-ls requires no filenames as arguments")

    if "-" in args.files:
        if len(args.files) > 1:
            parser.error("stdin can only be used with one file at a time")
        if args.only is None:
            parser.error("stdin mode requires '--only' to be set")

    success = driver_main(args)

    if not success:
        sys.exit(1)

    if args.verbosity:
        print("ok")


def list_codes() -> None:
    first = True
    for code in CODE_MAP.values():
        if code.hidden:
            continue
        if first:
            first = False
        else:
            print()

        if code.default_disabled:
            description: str = f"{code.message} (disabled by default)"
        else:
            description = code.message
        print(f"{code.code}: {description}")
        print(textwrap.indent(code.example, "    "))



from __future__ import annotations

import ast
import typing as t

from slyp.hashable_file import HashableFile

from ._base import ErrorRecordingVisitor
from .matching_branches import FindEquivalentBranchesVisitor

_VISITORS: t.Sequence[ErrorRecordingVisitor] = [
    FindEquivalentBranchesVisitor(),
]


def run_ast_checkers(file_obj: HashableFile) -> set[tuple[int, str]]:
    try:
        tree = ast.parse(file_obj.binary_content, filename=file_obj.filename)
    except SyntaxError:
        return {(0, "X001")}

    for visitor in _VISITORS:
        visitor.filename = file_obj.filename
        visitor.visit(tree)
    return {
        (lineno, code)
        for visitor in _VISITORS
        for (lineno, error_filename, code) in visitor.errors
        if error_filename == file_obj.filename
    }


def _clear_visitor_errors() -> None:
    for visitor in _VISITORS:
        visitor.errors = set()



from __future__ import annotations

import ast
import itertools
import typing as t

from ._base import ErrorRecordingVisitor


class CompareResult:
    __slots__ = ("matches", "is_trivial", "distance")

    def __init__(self, matches: bool, is_trivial: bool | None) -> None:
        self.matches = matches
        self.is_trivial = is_trivial
        self.distance = 0

    def __bool__(self) -> bool:
        return self.matches

    @property
    def adjacent(self) -> bool:
        return self.distance < 2


def is_trivial(
    node: ast.AST | list[ast.AST] | list[ast.expr] | list[ast.stmt] | None,
) -> bool:
    # if a "None" is reached during recursive descent, it means that a parent node
    # asked if one of its child elements is trivial, and there was no such element
    if node is None:
        return True
    # lists in AST are trivial if they are length 1 or 0
    # if they are length 1, their content must be trivial as well
    elif isinstance(node, list):
        if len(node) == 0:
            return True
        elif len(node) != 1:
            return False
        return is_trivial(node[0])
    # always-trivial things
    elif isinstance(node, (ast.Pass, ast.Break, ast.Continue, ast.Constant, ast.Name)):
        return True
    # trivial statements if their content is trivial
    elif isinstance(node, (ast.Return, ast.Yield)):
        return is_trivial(node.value)
    elif isinstance(node, ast.Expression):
        return is_trivial(node.body)
    elif isinstance(node, ast.Delete):
        return is_trivial(node.targets)
    elif isinstance(node, ast.Raise):
        return node.exc is None and node.cause is None
    # collections are trivial if they are empty or contain exactly one trivial item
    # this nicely matches a list check on the elements
    elif isinstance(node, (ast.List, ast.Tuple, ast.Set)):
        return is_trivial(node.elts)
    elif isinstance(node, ast.Dict):  # except for dict; that one has to be empty
        return len(node.keys) == 0
    # function calls are trivial if they have no keyword arguments and the argument
    # list is trivial but does not consist of nested function calls
    # i.e. this is trivial:
    #       foo(bar.baz)
    # but this is not:
    #       foo(bar(baz))
    # this is also not (nontrivial func node):
    #       foo(bar)(baz)
    elif isinstance(node, ast.Call):
        return is_trivial(node.func) and (
            node.keywords == []
            and (
                len(node.args) == 0
                or (
                    len(node.args) == 1
                    and not isinstance(node.args[0], ast.Call)
                    and is_trivial(node.args)
                )
            )
        )
    # attribute access is trivial if the value is trivial
    # i.e. in 'foo.bar', 'foo' must be trivial
    elif isinstance(node, ast.Attribute):
        return is_trivial(node.value)
    # assignments are trivial if there is only one target (not unpacking or chained) and
    # the value is trivial
    # (this intentionally does not cover AnnAssign)
    elif isinstance(node, ast.Assign):
        return (
            len(node.targets) == 1
            and isinstance(node.targets[0], ast.Name)
            and is_trivial(node.value)
        )
    return False


def compare_ast(
    left: ast.AST | list[ast.stmt], right: ast.AST | list[ast.stmt]
) -> CompareResult:
    if type(left) != type(right):  # noqa: E721
        return CompareResult(False, None)
    if isinstance(left, ast.AST):
        if not isinstance(right, ast.AST):
            raise NotImplementedError(
                "unreachable failure on mismatched AST shape compare!"
            )
        for lfield, rfield in itertools.zip_longest(
            ast.iter_fields(left), ast.iter_fields(right)
        ):
            lname, lvalues = lfield
            rname, rvalues = rfield
            if lname == rname and compare_ast(lvalues, rvalues):
                continue
            return CompareResult(False, None)
        return CompareResult(True, is_trivial(left))
    elif isinstance(left, list):
        if not isinstance(right, list):
            raise NotImplementedError(
                "unreachable failure on mismatched AST shape compare!"
            )
        for lvalue, rvalue in itertools.zip_longest(left, right):
            if compare_ast(lvalue, rvalue):
                continue
            return CompareResult(False, None)
        return CompareResult(True, is_trivial(left))
    else:
        res = left == right
        return CompareResult(res, is_trivial(left) if res else None)


def product_compare_ast(
    nodelist: t.Sequence[ast.AST | list[ast.stmt]],
) -> CompareResult:
    for index, item1 in enumerate(nodelist):
        for index2, item2 in enumerate(nodelist[index + 1 :]):
            if r := compare_ast(item1, item2):
                r.distance = index2 - index + 1
                return r
    return CompareResult(False, None)


def result_to_code(result: CompareResult) -> str:
    if result.adjacent:
        if result.is_trivial:
            return "W201"
        else:
            return "W200"
    else:
        if result.is_trivial:
            return "W203"
        else:
            return "W202"


class FindEquivalentBranchesVisitor(ErrorRecordingVisitor):
    # open questions:
    #
    # add support for for-else?
    #   async for-else?
    # add support for while-else?
    #   async while-else?
    #
    # others? what about bool ops like `foo() or foo()`?
    def _record(self, node: ast.AST, result: CompareResult) -> None:
        self.errors.add((node.lineno, self.filename, result_to_code(result)))

    def visit_Try(self, node: ast.Try) -> None:
        all_body_nodes: list[list[ast.stmt]] = [node.body]
        if node.handlers:
            all_body_nodes.extend(h.body for h in node.handlers)
        if node.orelse:
            all_body_nodes.append(node.orelse)
        if node.finalbody:
            all_body_nodes.append(node.finalbody)

        if r := product_compare_ast(all_body_nodes):
            self.errors.add((node.lineno, self.filename, result_to_code(r)))
        else:
            self.generic_visit(node)

    def visit_IfExp(self, node: ast.IfExp) -> None:
        if r := product_compare_ast([node.body, node.orelse]):
            self._record(node, r)
        else:
            self.generic_visit(node)

    def visit_If(self, node: ast.If) -> None:
        collected_branches = [node.body]

        current: list[ast.stmt] = node.orelse
        while len(current) == 1 and isinstance(current[0], ast.If):
            inner_if = current[0]
            collected_branches.append(inner_if.body)
            current = inner_if.orelse
        collected_branches.append(current)
        terminal_else_branch = current  # rename for clarity below

        if r := product_compare_ast(collected_branches):
            self._record(node, r)
        else:
            self.generic_visit(node.test)
            for subnode in node.body:
                self.generic_visit(subnode)
            for subnode in terminal_else_branch:
                self.generic_visit(subnode)



from __future__ import annotations

import dataclasses
import hashlib
import sys


@dataclasses.dataclass
class HashableFile:
    filename: str
    _sha: str | None = None
    _binary_content: bytes | None = None

    @property
    def binary_content(self) -> bytes:
        if self.is_stdio:
            if self._binary_content is None:
                self._binary_content = sys.stdin.buffer.read()
        elif self._binary_content is None:
            with open(self.filename, "rb") as fp:
                self._binary_content = fp.read()
        return self._binary_content

    @property
    def sha(self) -> str:
        if self._sha is None:
            self._sha = hashlib.sha256(self.binary_content).hexdigest()
        return self._sha

    def write(self, content: bytes) -> None:
        if self.is_stdio:
            sys.stdout.buffer.write(content)
            return

        with open(self.filename, "wb") as fp:
            fp.write(content)

        self._sha = None
        self._binary_content = content

    @property
    def is_stdio(self) -> bool:
        return self.filename == "-"



def test_check_captures_e100(check_text):
    res = check_text(
        """\
        x = "foo" "bar"
        """,
        filename="foo.py",
    )
    assert not res.success
    assert "foo.py:1: unnecessary string concat (E100)" in res.message_strings


def test_check_captures_e101(check_text):
    res = check_text(
        """\
        x = "foo" + "bar"
        """,
        filename="foo.py",
    )
    assert not res.success
    assert "foo.py:1: unnecessary string concat with plus (E101)" in res.message_strings


def test_check_captures_e101_multiline(check_text):
    res = check_text(
        """\
        x = '''foo
        ''' + "bar"
        """,
        filename="foo.py",
    )
    assert not res.success
    assert "foo.py:2: unnecessary string concat with plus (E101)" in res.message_strings



from __future__ import annotations

import dataclasses


@dataclasses.dataclass
class Result:
    messages: list[Message]
    success: bool

    @property
    def message_strings(self) -> list[str]:
        return [m.message for m in self.messages]

    def join(self, other: Result) -> Result:
        return Result(
            messages=self.messages + other.messages,
            success=self.success and other.success,
        )


@dataclasses.dataclass
class Message:
    message: str
    verbosity: int = 0
    priority: int = 1

    def __lt__(self, other: Message) -> bool:
        if self.priority != other.priority:
            return self.priority < other.priority
        return self.message < other.message



def test_check_passes_on_example_file(check_text):
    res = check_text(
        """
        x = "foo bar"

        foo(x="foo bar")
        foo(
            x=(
                "foo "
                "bar"
            )
        )

        y = {
            "foo": "bar",
        }
        y = {
            "foo": ("bar"
            "baz"),
        }

        z = (
            "alpha",
            "beta",
            "gamma",
        )
        z = [
            "alpha",
            "beta",
            "gamma",
        ]
        z = {
            "alpha",
            "beta",
            "gamma",
        }
        """
    )
    assert res.success


def test_can_disable_code_with_comment(check_text):
    res = check_text(
        """\
        def foo():
            x = "a " "b"  # slyp: disable=E100
        """,
        filename="foo.py",
    )
    assert res.success


def test_can_disable_code_via_category_disable(check_text):
    res = check_text(
        """\
        def foo():
            if bar():
                return baz("snork")
            elif qux():
                return quux()
            else:
                return baz("snork")
        """,
        filename="foo.py",
        disabled_codes={"W"},
    )
    assert res.success



import os
from unittest import mock

import pytest

from slyp.checkers import _clear_errors as _clear_checker_errors
from slyp.cli import main as cli_main
from slyp.driver import check_file, fix_file


@pytest.fixture(autouse=True)
def _auto_clear_checker_errors():
    _clear_checker_errors()


@pytest.fixture(autouse=True)
def _mock_parallel_processing():
    def fake_apply_async(func, args):
        mock_future = mock.Mock()
        mock_future.get.return_value = func(*args)
        return mock_future

    mock_pool = mock.Mock()
    mock_pool.apply_async = fake_apply_async
    with mock.patch("multiprocessing.pool.Pool", return_value=mock_pool):
        yield


@pytest.fixture
def run_cli(capsys):
    def _run_cli(args, assert_exit_code=0):
        with mock.patch("sys.argv", ["slyp"] + args):
            retcode = 0
            try:
                cli_main()
            except SystemExit as e:
                retcode = e.code
            if assert_exit_code is not None:
                assert retcode == assert_exit_code
            return retcode

    return _run_cli


def test_cli_invocation_simple(run_cli, tmpdir, capsys):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write('x = "foo bar"\n')
    with mock.patch(
        "slyp.driver.check_file", wraps=check_file
    ) as mock_check_file, mock.patch(
        "slyp.driver.fix_file", wraps=fix_file
    ) as mock_fix_file:
        assert run_cli(["foo.py"]) == 0, capsys.readouterr().out
        assert mock_check_file.call_count == 1
        assert mock_fix_file.call_count == 1


def test_double_cli_invocation_hits_cache(run_cli, tmpdir):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write('x = "foo bar"\n')
    with mock.patch(
        "slyp.driver.check_file", wraps=check_file
    ) as mock_check_file, mock.patch(
        "slyp.driver.fix_file", wraps=fix_file
    ) as mock_fix_file:
        run_cli(["foo.py"])
        assert mock_check_file.call_count == 1
        assert mock_fix_file.call_count == 1

        # running again does not call the checkers or fixers again!
        run_cli(["foo.py"])
        assert mock_check_file.call_count == 1
        assert mock_fix_file.call_count == 1


@pytest.mark.parametrize("cache_on_first_run", (True, False))
def test_double_cli_invocation_skips_cache_with_no_cache_flag(
    run_cli, tmpdir, cache_on_first_run
):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write('x = "foo bar"\n')
    with mock.patch(
        "slyp.driver.check_file", wraps=check_file
    ) as mock_check_file, mock.patch(
        "slyp.driver.fix_file", wraps=fix_file
    ) as mock_fix_file:
        run_cli(["foo.py"] if cache_on_first_run else ["--no-cache", "foo.py"])
        assert mock_check_file.call_count == 1
        assert mock_fix_file.call_count == 1

        # running again does all the work again
        run_cli(["--no-cache", "foo.py"])
        assert mock_check_file.call_count == 2
        assert mock_fix_file.call_count == 2


def test_cache_is_not_populated_under_no_cache(run_cli, tmpdir):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write('x = "foo bar"\n')
    with mock.patch(
        "slyp.driver.check_file", wraps=check_file
    ) as mock_check_file, mock.patch(
        "slyp.driver.fix_file", wraps=fix_file
    ) as mock_fix_file:
        # setup: run with `--no-cache`, which should *not* populate the cache
        run_cli(["--no-cache", "foo.py"])
        assert mock_check_file.call_count == 1
        assert mock_fix_file.call_count == 1

        # now run without `--no-cache` and confirm that there was no cache hit
        run_cli(["foo.py"])
        assert mock_check_file.call_count == 2
        assert mock_fix_file.call_count == 2



def test_check_captures_w200(check_text):
    res = check_text(
        """\
        def foo():
            if bar():
                return baz(quux("snork"))
            else:
                return baz(quux("snork"))
        """,
        filename="foo.py",
    )

    assert not res.success
    assert (
        "foo.py:2: two AST branches have identical contents (W200)"
        in res.message_strings
    )


def test_check_captures_w201(check_text):
    res = check_text(
        """\
        def foo():
            if bar():
                return baz()
            else:
                return baz()
        """,
        filename="foo.py",
    )

    assert not res.success
    assert (
        "foo.py:2: two AST branches have identical trivial contents (W201)"
        in res.message_strings
    )


def test_check_captures_w202(check_text):
    res = check_text(
        """\
        def foo():
            if bar():
                return baz("snork", flip="flop")
            elif qux():
                return quux()
            else:
                return baz("snork", flip="flop")
        """,
        filename="foo.py",
    )

    assert not res.success
    assert (
        "foo.py:2: two non-adjacent AST branches have identical contents (W202)"
        in res.message_strings
    )


def test_check_captures_w203(check_text):
    res = check_text(
        """\
        def foo():
            if bar():
                return baz("snork")
            elif qux():
                return quux()
            else:
                return baz("snork")
        """,
        filename="foo.py",
    )

    assert not res.success
    assert (
        "foo.py:2: two non-adjacent AST branches have identical trivial contents (W203)"
        in res.message_strings
    )



from __future__ import annotations

import libcst
import libcst.matchers

from ._base import ErrorCollectingVisitor

# SimpleWhitespace defines whitespace as seen between tokens in most contexts
# it can contain newlines *if* there is a preceding backslash escape
SIMPLE_WHITESPACE_NO_NEWLINE_MATCHER = libcst.matchers.SimpleWhitespace(
    value=libcst.matchers.MatchIfTrue(lambda x: "\n" not in x)
)


class StrConcatErrorCollector(ErrorCollectingVisitor):
    METADATA_DEPENDENCIES = (libcst.metadata.PositionProvider,)

    def visit_ConcatenatedString(self, node: libcst.ConcatenatedString) -> None:
        # check for 'unnecessary string concat' situations
        # e.g.
        #   x = "foo " "bar"
        #
        # these are easily introduced when strings change in length and `black` is run
        # also common when `black` runs for the first time on a project
        if libcst.matchers.matches(
            node.whitespace_between, SIMPLE_WHITESPACE_NO_NEWLINE_MATCHER
        ):
            lpos = self.get_metadata(libcst.metadata.PositionProvider, node.left).start
            self.errors.add((lpos.line, self.filename, "E100"))

    def visit_BinaryOperation(self, node: libcst.BinaryOperation) -> None:
        # check for 'unnecessary string concat' situations with explicit `+`
        # e.g.
        #   x = "foo " + "bar"
        if libcst.matchers.matches(
            node,
            libcst.matchers.BinaryOperation(
                operator=libcst.matchers.Add(),
                left=libcst.matchers.OneOf(
                    libcst.matchers.SimpleString(),
                    libcst.matchers.ConcatenatedString(),
                    libcst.matchers.FormattedString(),
                ),
                right=libcst.matchers.OneOf(
                    libcst.matchers.SimpleString(),
                    libcst.matchers.ConcatenatedString(),
                    libcst.matchers.FormattedString(),
                ),
            ),
        ):
            lpos = self.get_metadata(libcst.metadata.PositionProvider, node.left).end
            rpos = self.get_metadata(libcst.metadata.PositionProvider, node.right).start
            if lpos.line == rpos.line:
                self.errors.add((lpos.line, self.filename, "E101"))
